package com.analogbit.puzzletools.client.slidingblocks.gui;

import java.util.HashMap;
import java.util.Map.Entry;

import com.allen_sauer.gwt.dnd.client.DragContext;
import com.allen_sauer.gwt.dnd.client.DragEndEvent;
import com.allen_sauer.gwt.dnd.client.DragHandler;
import com.allen_sauer.gwt.dnd.client.DragStartEvent;
import com.allen_sauer.gwt.dnd.client.VetoDragException;
import com.allen_sauer.gwt.dnd.client.drop.GridConstrainedDropController;
import com.analogbit.puzzletools.client.slidingblocks.state.Block;
import com.analogbit.puzzletools.client.slidingblocks.state.BlockInPositionGoalState;
import com.analogbit.puzzletools.client.slidingblocks.state.GoalState;
import com.analogbit.puzzletools.client.slidingblocks.state.Grid;
import com.analogbit.puzzletools.client.slidingblocks.state.MultiGoalState;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.Button;
import com.google.gwt.user.client.ui.HasHorizontalAlignment;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Label;

/**
 * Wizard panel that contains the initial state on the left half and the desired state on the right.
 */
public class KeyPanel extends WizardPanel {

	private KeySourceGridDisplay sourceGrid;
	private KeyDestGridDisplay destGrid;
	private Button resetButton;
	private AbsolutePanel outer;
	private Grid grid;
	private Grid keyGrid;
	private HashMap<Block, Block> blockMapping;
	
	public KeyPanel() {
		super("Setup Key Blocks", "The solver needs to know which blocks are important and need to be moved. Drag blocks from the Initial State grid to the Goal State grid based on the location where you want them to end up.");
		
		outer = new AbsolutePanel();
		sourceGrid = new KeySourceGridDisplay(outer);
		destGrid = new KeyDestGridDisplay();
		resetButton = new Button("Reset Goal State");
		blockMapping = new HashMap<Block, Block>();
		resetButton.addClickHandler(new ClickHandler() {
			public void onClick(ClickEvent event) {
				setGrid(grid);
			}
		});
		
		HorizontalPanel labelPanel = new HorizontalPanel();
		labelPanel.setWidth("100%");
		labelPanel.setHorizontalAlignment(HasHorizontalAlignment.ALIGN_CENTER);
		labelPanel.add(new Label("Initial State"));
		labelPanel.add(new Label("Goal State"));
		
		outer.add(sourceGrid);
		outer.add(destGrid);
		add(outer);
		add(labelPanel);
		add(resetButton);		
	
		sourceGrid.getDragController().registerDropController(new GridConstrainedDropController(destGrid, GridBlock.CELL_SIZE, GridBlock.CELL_SIZE));
		sourceGrid.getDragController().addDragHandler(new KeyDropDragHandler());
		destGrid.getDragController().registerDropController(new GridConstrainedDropController(destGrid, GridBlock.CELL_SIZE, GridBlock.CELL_SIZE));
		destGrid.getDragController().addDragHandler(new KeyMoveDragHandler());

	}
	
	public Grid getGrid() {
		return grid;
	}
	
	public void setGrid(Grid grid) {
		this.grid = grid;
		this.sourceGrid.setGridDisplaying(grid);
		this.keyGrid = grid.createShallowImmobilityGrid(true);
		this.destGrid.setGridDisplaying(keyGrid);
		this.blockMapping.clear();
		
		int width = grid.getWidth() * GridBlock.CELL_SIZE;
		int height = grid.getHeight() * GridBlock.CELL_SIZE;
		int MARGIN = 8*2;
		width += MARGIN;
		height += MARGIN;
		
		outer.setPixelSize(width*2, height);
		outer.setWidgetPosition(sourceGrid, 0, 0);
		outer.setWidgetPosition(destGrid, width, 0);
	}

	/**
	 * Creates the goal state object for the solution to the panel based on 
	 * what the user has dragged into place.
	 * @return The goal state desired by the user
	 */
	public GoalState getGoalState() {
		MultiGoalState g = new MultiGoalState();
		
		for (Entry<Block, Block> entry : blockMapping.entrySet()) {
			Block source = entry.getKey();
			Block dest = entry.getValue();
			
			if (source.isImmobile()) {
				continue;
			}
			g.add(new BlockInPositionGoalState(grid, source, dest.getLeft(), dest.getTop()));
		}
		return g;
	}
	
	/*private class KeyDropController extends GridConstrainedDropController {

		public KeyDropController(AbsolutePanel dropTarget) {
			super(dropTarget, GridBlock.CELL_SIZE, GridBlock.CELL_SIZE);
		}

		@Override
		public void drop(Widget widget, int left, int top) {
			super.drop(widget, left, top);
		}
		
		@Override
		public void onPreviewDrop(DragContext context) throws VetoDragException {
			GridBlock gb = (GridBlock)context.draggable;
			int x = (context.desiredDraggableX-getDropTarget().getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int y = (context.desiredDraggableY-getDropTarget().getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			
			Block sourceBlock = gb.getSourceBlock();
			Block clone = sourceBlock.cloneWithNewPos(x, y);
			if (keyGrid.addBlockIfLegal(clone)) {
				blockMapping.put(gb.getSourceBlock(), clone);
			} else {
				throw new VetoDragException();
			}
		}

		@Override
		public void onDrop(DragContext context) {
			super.onDrop(context);
			
			//make the draggable dissapear now
			context.draggable.removeFromParent();
		}
		
		
	}*/
	
	/**
	 * Responsible for dropping the blocks onto the key grid.
	 */
	private class KeyDropDragHandler implements DragHandler {

		public void onDragEnd(DragEndEvent event) {
			DragContext context = event.getContext();
			if (context.vetoException == null) {
				//make the draggable dissapear now
				context.draggable.removeFromParent();
			}
		}

		public void onDragStart(DragStartEvent event) {
		
		}

		public void onPreviewDragEnd(DragEndEvent event)
				throws VetoDragException {

			DragContext context = event.getContext();
			GridBlock gb = (GridBlock)context.draggable;
			
			//int x = (context.desiredDraggableX-getDropTarget().getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			//int y = (context.desiredDraggableY-getDropTarget().getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int x = (gb.getAbsoluteLeft() - destGrid.getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int y = (gb.getAbsoluteTop() - destGrid.getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			
			Block sourceBlock = gb.getSourceBlock();
			Block clone = sourceBlock.cloneWithNewPos(x, y);
			if (keyGrid.addBlockIfLegal(clone)) {
				blockMapping.put(gb.getSourceBlock(), clone);
			} else {
				throw new VetoDragException();
			}
			
		}

		public void onPreviewDragStart(DragStartEvent event)
				throws VetoDragException {	
		}
	}
	
	/**
	 * Responsible for moving a gridblock around on the key setup block.
	 */
	private class KeyMoveDragHandler implements DragHandler {
		public void onDragEnd(DragEndEvent event) {
			DragContext context = event.getContext();
			GridBlock gb = (GridBlock)context.draggable;
			destGrid.updateGridBlockPosition(gb);
		}

		public void onDragStart(DragStartEvent event) {		
		}

		public void onPreviewDragEnd(DragEndEvent event)
				throws VetoDragException {	
			
			DragContext context = event.getContext();
			GridBlock gb = (GridBlock)context.draggable;
			//int x = (context.desiredDraggableX-getDropTarget().getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			//int y = (context.desiredDraggableY-getDropTarget().getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int x = (gb.getAbsoluteLeft() - destGrid.getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int y = (gb.getAbsoluteTop() - destGrid.getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			
			if (!keyGrid.moveBlockIfLegal(gb.getSourceBlock(), x, y)) {
				throw new VetoDragException();
			}
		}

		public void onPreviewDragStart(DragStartEvent event)
				throws VetoDragException {	
		}
		
	}
		
}
